using System;
using System.Net.Sockets;
using System.IO;
using System.Text;
using System.Threading;
using System.Collections;


namespace Zimnox.Net.CTL
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	public class TOCConnection
	{
		private string loginServer = "aimexpress.oscar.aol.com";
		private int loginPort = 5190;
		private string authorizerServer = "login.oscar.aol.com";
		private int authorizerPort = 29999;
   
		private TcpClient connection = null;
		private NetworkStream stream = null;
		private Thread connectionThread = null;

		private string REVISION = "\"TIC://Revision: 1.61 \" 160 US \"\" \"\" 3 0 30303 -kentucky -utf8 94791632";

		private int seqNo = 4;
		private string VERSION = "CTL";
		private string username = "";
		private string password = "";
		private int permitMode = 1;
		
		private bool isConnected = false;

		private ASCIIEncoding Enc = new ASCIIEncoding();

		private Hashtable buddies = new Hashtable();
		private ArrayList denyBuddies = new ArrayList();
		private ArrayList permitBuddies = new ArrayList();
		private ArrayList listeners = new ArrayList(2);
		

		protected enum FrameType 
		{
			SignOn = 1,
			Data = 2,
			Error = 3,
			SignOff = 4,
			KeepAlive = 5
		}


		/// <summary>
		/// Default constructor, assigns default client name
		/// as "CTL". The client name is what the TOC servers
		/// use to distinguish different clients.
		/// </summary>
		public TOCConnection()
		{			
		}


		/// <summary>
		/// The client name is what the TOC servers
		/// use to distinguish different clients.
		/// </summary>
		/// <param name="clientName">name to use to indentify your client on the network</param>
		public TOCConnection(string clientName)
		{	
			VERSION = clientName;
		}



		/// <summary>
		/// port of the authorization server used for authentication
		/// </summary>
		public int AuthorizationPort
		{
			get
			{
				return authorizerPort;
			}
			set
			{
				authorizerPort = value;
			}
		}



		/// <summary>
		/// address of the authorization server used for authentication
		/// </summary>
		public string AuthorizationServer
		{
			get
			{
				return authorizerServer;
			}
			set
			{
				authorizerServer = value;
			}
		}



		/// <summary>
		/// 
		/// </summary>
		/// <returns>username used to login in with for this connection</returns>
		public string GetUsername()
		{
			return username;
		}



		/// <summary>
		/// port of the login server used for connection
		/// </summary>
		public int LoginPort
		{
			get
			{
				return loginPort;
			}
			set
			{
				loginPort = value;
			}
		}



		/// <summary>
		/// address of the login server used for the connection
		/// this is the address over which the communication happens
		/// </summary>
		public string LoginServer
		{
			get
			{
				return loginServer;
			}
			set
			{
				loginServer = value;
			}
		}



		/// <summary>
		/// the thread function, listens for incoming data
		/// </summary>
		public void run()
		{
			while(isConnected)
			{
				try
				{
					string flap = readFlap();
					if (flap != null)
						parseTocData(flap);
				}
				catch(Exception e)
				{
					System.Console.WriteLine(e);
				}
			}

			closeConnection();
		}



		/// <summary>
		/// next sequence number for the next packet
		/// </summary>
		/// <returns>FLAP sequence number</returns>
		private int nextSeqNo()
		{
			seqNo = (seqNo + 1) & 65535;
			return seqNo;
		}


		/// <summary>
		/// Parses the data received from TOC server 
		/// and fires the appropriate events
		/// </summary>
		/// <param name="data">the data FLAP to parse</param>
		private void parseTocData(string data)
		{
			System.Console.WriteLine(data);

			
			char[] delim = {':'};
			String[] tokens = data.Split(delim);

			//configuration file, contains buddy names and groups
			if (tokens[0].Equals("CONFIG2"))
			{
				string tmp = "";
				for(int i=1; i<tokens.Length; i++)
				{
					tmp += ":"+tokens[i];
				}
				processConfig(tmp);
				fireConfigReceived();
				return;
			}


			//incoming message
			//DEPRECATED due to new login method, kept for compatibility reasons
			if (tokens[0].Equals("IM_IN2"))
			{
				if (tokens.Length >= 4)
				{
					string buddyName	= tokens[1];
					string auto			= tokens[2];
					string message		= tokens[3];

					for(int i=4; i<tokens.GetLength(0); i++)
						message += ":"+tokens[i];
					
					TOCBuddy buddy = GetBuddy(buddyName);
					bool isBuddy = true;
	            
					//buddy is not on list
					if (buddy == null)
					{
						buddy = new TOCBuddy(buddyName, null);
						isBuddy = false;
					}

					fireMessageReceived(buddyName, message.Trim(), isBuddy, bool.Parse(auto));
				}
				else
				{
					Console.WriteLine("IM_IN2: Expected >= 4 but found "+tokens.GetLength(0));
				}
				return;
			}


			//incoming message
			if (tokens[0].Equals("IM_IN_ENC2"))
			{
				if (tokens.GetLength(0) >= 10)
				{
					string buddyName    = tokens[1];
					string auto         = tokens[2];
					string unknown1     = tokens[3];
					string unknown2     = tokens[4];
					string status       = tokens[5];
					string unknown3     = tokens[6];
					string unknown4     = tokens[7];
					string language     = tokens[8];
					string message      = tokens[9];

					for(int i=10; i<tokens.GetLength(0); i++)
						message += ":"+tokens[i];

					TOCBuddy buddy = GetBuddy(buddyName);
					bool isBuddy = true;

					//buddy is not on list
					if (buddy == null)
					{
						buddy = new TOCBuddy(buddyName, null);
						isBuddy = false;
					}
					fireMessageReceived(buddyName, message.Trim(), isBuddy, bool.Parse(auto), language);
				}
				else
				{
					Console.WriteLine("IM_IN_ENC2: Expected >= 10 but found "+tokens.Length);
				}
				return;
			}

			//typing notifications
			if (tokens[0].Equals("CLIENT_EVENT2"))
			{
				if (tokens.Length == 3)
				{
					string buddy	= tokens[1];
					int status		= int.Parse(tokens[2]);
			
					fireClientEvent(buddy, status);
				}
				else
				{
					Console.WriteLine("CLIENT_EVENT2: Expected 3 but found "+tokens.Length);
				}
				return;
			}        

			//possibly used for buddy art/icon
			if (tokens[0].Equals("BART2"))
			{
				if (tokens.Length >= 3)
				{
					string buddy = tokens[1];
					string unknown = "";

					for(int i=2; i<tokens.Length; i++)
						unknown += ":"+tokens[i];
	            
					fireBart(buddy, unknown);
				}
				else
				{
					Console.WriteLine("BART2: Expected >= 3 but found "+tokens.Length);
				}
				return;
			}

			//supposedly sent by server upon successfuly addition of a new buddy
			//this command is unverified
			if (tokens[0].Equals("NEW_BUDDY_REPLY2"))
			{
				if (tokens.Length == 3)
				{
					string buddyName	= tokens[1];
					string action		= tokens[2];
					
					if (action.ToLower().Equals("added"))
					{
						fireBuddyAdded(buddyName);
					}
				}
				else
				{
					Console.WriteLine(": Expected 3 but found "+tokens.Length);
				}
				return;
			}

			//these are supposedly returned when the buddy list is changed
			//from another location/client
			if (tokens[0].Equals("INSERTED2"))
			{
				if (tokens[1].Equals("g")) //new group added to list
				{
					//fireGroupAdded(tokens[2]);
				}
				else if (tokens[1].Equals("b")) //new buddy added to list
				{
					string alias		= tokens[2];
					string buddyName	= tokens[3];
					string group		= tokens[4];

					TOCBuddy buddy = new TOCBuddy(buddyName, group);
					buddies.Add(normalizeString(buddyName), buddy);
					fireBuddyAdded(buddyName, group);
				}
				else if (tokens[1].Equals("d")) //name added to deny list
				{
					string buddyName = tokens[2];
					denyBuddies.Add(buddyName);
					fireDenyAdded(buddyName);
				}
				else if (tokens[1].Equals("p")) //name added to permit list
				{
					string buddyName = tokens[2];
					permitBuddies.Add(buddyName);
					firePermitAdded(buddyName);
				}
				return;
			}

			//update buddy status online
			if (tokens[0].Equals("UPDATE_BUDDY2"))
			{
				if (tokens.Length == 8)
				{
					string buddyName	= tokens[1];
					bool onlineStatus	= bool.Parse(tokens[2]);
					int warning			= int.Parse(tokens[3]);
					long signonTime		= long.Parse(tokens[4]);
					int idleTime		= int.Parse(tokens[5]);
					string uc			= tokens[6];
					string unknown		= tokens[7];

					TOCBuddy buddy = GetBuddy(buddyName);

					if (onlineStatus)
					{
						buddy.SetIdle(idleTime);
						buddy.SetSignonTime(signonTime);

						//if buddy isn't already online
						if (!buddy.IsOnline())
						{
							buddy.SetOnline(true);
							fireBuddySignedOn(buddyName);
						}

						if (uc.Length == 3)
						{
							if (uc[2] == 'U')
							{
								if (!buddy.IsAway()) // if buddy is not already set away
								{
									buddy.SetAway(true);
									fireBuddyAway(buddyName);
								}
							}
						}
						else
						{
							if (buddy.IsAway())
							{
								buddy.SetAway(false);
								fireBuddyReturned(buddyName);
							}
						}

						if (uc[0] == 'A' || uc[1] == 'A')
							buddy.Service = TOCConstants.SERVICE_AOL;
						if (uc[1] == 'C')
							buddy.Service = TOCConstants.SERVICE_CELL;
						if (uc[1] == 'U')
							buddy.Service = TOCConstants.SERVICE_UNCONFIRMED;
						if (uc[1] == 'O')
							buddy.Service = TOCConstants.SERVICE_NORMAL;
						if (uc[1] == 'I')
							buddy.Service = TOCConstants.SERVICE_ICQ;
					}
					else
					{
						if (buddy.IsOnline())
						{
							buddy.SetOnline(false);
							fireBuddySignedOff(buddyName);
						}
					}

				}
				else
				{
					Console.WriteLine("UPDATE_BUDDY2: Expected 8 but found "+tokens.Length);
				}
				return;
			}

			//error number with possible extra arguments
			if (tokens[0].Equals("ERROR"))
			{
				int error = int.Parse(tokens[1]);
				string args = "";
				
				if (tokens.Length > 2)
				{
					args = tokens[2];
					for(int i=3; i<tokens.Length; i++)
						args += ":"+tokens[i];
				}

				fireTocError(error, args.Split(':'));
				return;
			}

			//warned by another user
			//buddy name is blank if it was anonymous
			if (tokens[0].Equals("EVILED"))
			{
				if (tokens.Length == 3)
				{
					int amount = int.Parse(tokens[1]);
					string buddyName = tokens[2];

					fireWarning(buddyName, amount);
				}
				else
				{
					Console.WriteLine("EVILED: Expected 3 but found "+tokens.Length);
				}
				return;
			}

			//successfully joined a chatroom
			if (tokens[0].Equals("CHAT_JOIN"))
			{
				if (tokens.Length == 3)
				{
					string roomID	= tokens[1];
					string roomName = tokens[2];
					fireChatJoined(roomID, roomName);
				}
				else
				{
					Console.WriteLine("CHAT_JOIN: Expected 3 but found "+tokens.Length);
				}
				return;
			}

			//incoming chatroom message
			//DEPRECATED due to new login method, CHAT_IN_ENC2 is now used
			if (tokens[0].Equals("CHAT_IN"))
			{
				if (tokens.Length >= 5)
				{
					string roomID	= tokens[1];
					string sender	= tokens[2];
					bool whisper	= bool.Parse(tokens[3]);
					string message	= tokens[4];
					for(int i=5; i<tokens.Length; i++)
						message += ":"+tokens[i];

					fireChatMessageReceived(roomID, sender, message, whisper);
				}
				else
				{
					Console.WriteLine("CHAT_IN: Expected >= 5 but found "+tokens.Length);
				}
				return;
			}

			//incoming chatroom message
			if (tokens[0].Equals("CHAT_IN_ENC2"))
			{
				if (tokens.Length >= 7)
				{
					string roomID	= tokens[1];
					string sender	= tokens[2];
					bool whisper	= bool.Parse(tokens[3]);
					string unknown	= tokens[4];
					string language = tokens[5];
					string message	= tokens[6];
					for(int i=7; i<tokens.Length; i++)
						message += ":"+tokens[i];

					fireChatMessageReceived(roomID, sender, message, whisper, language, unknown);
				}
				else
				{
					Console.WriteLine("CHAT_IN_ENC2: Expected >= 7 but found "+tokens.Length);
				}
				return;
			}

			//update status of a buddy in a chatroom
			//command can return multiple buddy names at once
			if (tokens[0].Equals("CHAT_UPDATE_BUDDY"))
			{
				if (tokens.Length >= 3)
				{
					string roomID	= tokens[1];
					bool inside		= bool.Parse(tokens[2]);
					for(int i=3; i<tokens.Length; i++){
						fireChatUpdate(roomID, inside, tokens[i]);
					}
				}
				else
				{
					Console.WriteLine("CHAT_UPDATE_BUDDY: Expected >= 3 but found "+tokens.Length);
				}
				return;
			}

			//invitation to join a chatroom
			if (tokens[0].Equals("CHAT_INVITE"))
			{
				if (tokens.Length >= 5)
				{
					string roomName	= tokens[1];
					string roomID	= tokens[2];
					string sender	= tokens[3];
					string message	= tokens[4]; 
					for(int i=5; i<tokens.Length; i++)
						message += ":"+tokens[i];

					fireChatInvite(roomID, roomName, sender, message);
				}
				else
				{
					Console.WriteLine("CHAT_INVITE: Expected >= 5 but found "+tokens.Length);
				}
				return;
			}

			//dropped out of chatroom
			if (tokens[0].Equals("CHAT_LEFT"))
			{
				if (tokens.Length == 2)
				{
					string roomID = tokens[1];
					fireChatDropped(roomID);
				}
				else
				{
					Console.WriteLine("CHAT_LEFT: Expected 2 but found "+tokens.Length);
				}
				return;
			}

			//received when user sends "toc_get_info"
			if (tokens[0].Equals("GOTO_URL"))
			{
				if (tokens.Length == 3)
				{
					fireUrlReceived(tokens[2]);
				}
				else
				{
					Console.WriteLine("GOTO_URL: Expected 3 but found "+tokens.Length);
				}
				return;
			}

			//
			if (tokens[0].Equals("DIR_STATUS"))
			{
				if (tokens.Length >= 3)
				{
					int code = int.Parse(tokens[1]);
					string args = "";
					for(int i=2; i<tokens.Length; i++)
						args += ":"+tokens[i];

					fireDirStatus(code, args.Split(':'));
				}
				else
				{
					Console.WriteLine("DIR_STATUS: Expected >= 3 but found "+tokens.Length);
				}
				return;
			}

			//
			if (tokens[0].Equals("ADMIN_NICK_STATUS"))
			{
				if (tokens.Length >= 2)
				{
					int code = int.Parse(tokens[1]);
					string args = "";
					for(int i=2; i<tokens.Length; i++)
						args += ":"+tokens[i];

					fireNickStatus(code, args.Split(':'));
				}
				else
				{
					Console.WriteLine("ADMIN_NICK_STATUS: Expected >= 2 but found "+tokens.Length);
				}
				return;
			}

			//
			if (tokens[0].Equals("ADMIN_PASSWD_STATUS"))
			{
				if (tokens.Length >= 2)
				{
					int code = int.Parse(tokens[1]);
					string args = "";
					for(int i=2; i<tokens.Length; i++)
						args += ":"+tokens[i];

					firePasswordStatus(code, args.Split(':'));
				}
				else
				{
					Console.WriteLine("ADMIN_PASSWD_STATUS: Expected >= 2 but found "+tokens.Length);
				}
				return;
			}

			//server has paused communication and any commands sent to server
			//after this point will be ignored until the server has sent
			//the signon command.
			if (tokens[0].Equals("PAUSE"))
			{
				if (tokens.Length == 1)
				{
					firePause(true);
				}
				else
				{
					Console.WriteLine("PAUSE: Expected 1 but found "+tokens.Length);
				}
				return;
			}

			//server has resumed communication
			if (data.Equals("SIGN_ON:TOC2.0"))
			{
				firePause(false);
				return;
			}

			//
			if (tokens[0].Equals("RVOUS_PROPOSE"))
			{
				if (tokens.Length >= 9)
				{
					string buddyName	= tokens[1];
					string UUID			= tokens[2];
					string cookie		= tokens[3];
					int seq				= int.Parse(tokens[4]);
					string rip			= tokens[5];
					string pip			= tokens[6];
					string vip			= tokens[7];
					string port			= tokens[8];
					string tlv			= "";
					for(int i=9; i<tokens.Length; i++)
						tlv += ":"+tokens[i];
 
					fireRvousPropose(buddyName, UUID, cookie, seq, rip, pip, vip, port, tlv.Split(':'));
				}
				else
				{
					Console.WriteLine(": Expected >= 9 but found "+tokens.Length);
				}
				return;
			}

			//
			if (tokens[0].Equals("DELETED2"))
			{
				string type = tokens[1];
				
				if (type.Equals("b")) //buddy was deleted
				{
					string buddyName  = tokens[2];
					string buddyGroup = tokens[3];
					buddies.Remove(normalizeString(buddyName));
					fireBuddyRemoved(buddyName, buddyGroup);
				}
				else if (type.Equals("d")) //buddy removed from deny list
				{
					string buddyName = tokens[2];
					denyBuddies.Remove(buddyName);
					fireBuddyRemovedFromDeny(buddyName);
				}
				else if (type.Equals("p")) //buddy removed from permit list
				{
					string buddyName = tokens[2];
					permitBuddies.Remove(buddyName);
					fireBuddyRemovedFromPermit(buddyName);
				}
				else if (type.Equals("g")) //group was deleted
				{
					string groupName = tokens[2];
					fireGroupRemoved(groupName);
				}
				return;
			}
		}



		/// <summary>
		/// initial sign on sends a config file structure
		/// which contains the current buddy list with
		/// associated groups
		/// </summary>
		/// <param name="config">config formatted string</param>
		private void processConfig(string config)
		{
			StringReader reader = new StringReader(config);
			string group = "";
			string line;
			TOCBuddy buddy = null;
			while((line = reader.ReadLine()) != null)
			{				
				string[] tokens = line.Split(new char[]{':'});

				if(tokens[0].Equals("done"))
					return;
				
				char type = line[0];
				switch(type)
				{
					case 'g':
						group = tokens[1];
						break;
					case 'b': 
						buddies.Add(normalizeString(tokens[1]),new TOCBuddy(tokens[1],group));
						break;
					case 'd': 						
						//buddy = new TOCBuddy(tokens[1], group);
						//buddy.SetPermissionState(TOCBuddy.PERMISSION_DENY);
						//buddies.Add(normalizeString(tokens[1]), buddy);
						denyBuddies.Add(tokens[1]);
						break;
					case 'p': 
						buddy = new TOCBuddy(tokens[1], group);
						buddy.SetPermissionState(TOCBuddy.PERMISSION_PERMIT);
						buddies.Add(normalizeString(tokens[1]), buddy);
						permitBuddies.Add(tokens[1]);
						break;
					case 'm': 
						SetPermitMode(int.Parse(tokens[1]));
						break;
				}
			}
		}



		/// <summary>
		/// returns the associated TOCBuddy object of the 
		/// passed in string buddy name. Buddy name does
		/// not have to be normalized, the normalization
		/// is called from within the function itself.
		/// </summary>
		/// <param name="buddyName">buddy's name</param>
		/// <returns>TOCBuddy object</returns>
		public TOCBuddy GetBuddy(string buddyName)
		{
			return (TOCBuddy)buddies[normalizeString(buddyName)];
		}


		/// <summary>
		/// Returns current buddy list as a Hashtable
		/// key->value = normalized buddy name -> TOCBuddy
		/// </summary>
		/// <returns>Hashtable</returns>
		public Hashtable GetBuddies()
		{
			return buddies;
		}


		/// <summary>
		/// Checks to see if buddy is on the permit list
		/// </summary>
		/// <param name="buddyName"></param>
		/// <returns>true/false</returns>
		public bool IsBuddyOnPermit(string buddyName)
		{
			return permitBuddies.Contains(buddyName);
		}


		/// <summary>
		/// Checks to see if buddy in on the deny/block list
		/// </summary>
		/// <param name="buddyName"></param>
		/// <returns>true/false</returns>
		public bool IsBuddyOnDeny(string buddyName)
		{
			return denyBuddies.Contains(buddyName);
		}


		/// <summary>
		/// See TOCConstants details about different modes
		/// </summary>
		/// <param name="mode">View the "service" constants in TOCConstants</param>
		/// <seealso cref="TOCConstants"/>
		public void SetPermitMode(int mode)
		{
			sendFlap((byte)FrameType.Data, "toc2_set_pdmode "+mode);
			permitMode = mode;
		}


		/// <summary>
		/// See TOCConstants details about different modes
		/// </summary>
		/// <returns>permit mode</returns>
		/// <seealso cref="TOCConstants"/>
		public int GetPermitMode()
		{
			return permitMode;
		}


		/// <summary>
		/// sends an IM to a buddy
		/// </summary>
		/// <param name="buddy">name of buddy to send message to</param>
		/// <param name="message">message text</param>
		public void SendMessage(string buddy, string message)
		{
			sendFlap((byte)FrameType.Data, "toc2_send_im "+normalizeString(buddy)+" \""+message+"\"");
		}


		/// <summary>
		/// sets the current user status to away
		/// </summary>
		/// <param name="message">away message</param>
		public void SetAway(string message)
		{
			sendFlap((byte)FrameType.Data, "toc_set_away \""+message+"\"");
		}


		/// <summary>
		/// return from away status
		/// alternatively, you can use SetAway("") and 
		/// simply pass an empty (but not null) string
		/// to also return from away status
		/// </summary>
		public void ReturnFromAway()
		{			
			sendFlap((byte)FrameType.Data, "toc_set_away "+"\0");
		}


		/// <summary>
		/// set the capabilities of the client
		/// see TOCConstants for available caps
		/// </summary>
		/// <param name="caps">array of capability IDs</param>
		/// <seealso cref="TOCConstants"/>
		public void SetCaps(string[] caps)
		{
			string c = caps[0];
			for(int i=1; i<caps.Length; i++)
				c = " "+c;
			sendFlap((byte)FrameType.Data, "toc_set_caps "+c+"\0");
		}


		/// <summary>
		/// Signs onto the TOC network using the TOC2.0 protocol
		/// </summary>
		/// <param name="user">username</param>
		/// <param name="pass">password</param>
		/// <seealso cref="SignOff()"/>
		public void SignOn(string user, string pass)
		{
			if (!isConnected)
			{
				if (user != null && pass != null)
				{
					this.username = user;
					this.password = pass;

					if (ConnectToTOC())
					{
						if (sendSignonFlap())
						{
							connectionThread = new Thread(new ThreadStart(run));
							connectionThread.Start();
						}
						else
							throw new IOException("Authorization error");
					}
					else
						throw new IOException("Error connecting to TOC server");
				}
			}
			else
				throw new IOException("You are already connected");   
		}


		/// <summary>
		/// true/false if connection is active or not
		/// </summary>
		/// <returns>true if connection is active</returns>
		public bool IsAlive()
		{
			return isConnected;
		}


		/// <summary>
		/// Attempts to establish a connection with TOC server
		/// </summary>
		/// <returns>true if connection was successful</returns>
		private bool ConnectToTOC()
		{
			connection = new TcpClient(loginServer, loginPort);

			stream = connection.GetStream();

			if (connection == null || stream == null)
				return false;

			return true;
		}


		/// <summary>
		/// Signals to disconnects from TOC server
		/// Any calls to GetUsername() after this will
		/// return null
		/// <seealso cref="SignOff()"/>
		/// </summary>
		public void SignOff()
		{
			username = null;
			password = null;
			isConnected = false;
		}


		/// <summary>
		/// Closes connection and input/output streams
		/// </summary>
		private void closeConnection()
		{
			if (stream != null)
			{
				stream.Close();
				stream = null;
			}
			if (connection != null)
			{
				connection.Close();
				connection = null;
			}
		}


		/// <summary>
		/// Sends signon FLAP
		/// </summary>
		/// <returns>true if authorization successful</returns>
		private bool sendSignonFlap()
		{
			try
			{
				string nUsername = normalizeString(username);

				writeString("FLAPON\r\n\r\n");
				stream.Flush();

				//6 byte header, 4 byte FLAP version
				byte[] signonBytes = new byte[10];
				//TOC sends Client FLAP SIGNON
				stream.Read(signonBytes,0,10);

				//Client sends TOC FLAP SIGNON
				/* FLAP Header */
				stream.WriteByte((byte)'*');
				stream.WriteByte((byte)FrameType.SignOn);
				writeShort((short)seqNo++);
				writeShort((short)(username.Length+8));
				/* SIGNON Frame */
				//FLAP version 1, integer value
				stream.WriteByte(0);
				stream.WriteByte(0);
				stream.WriteByte(0);
				stream.WriteByte(1);
				writeShort(1); //TLV Tag
				writeShort((short)nUsername.Length);
				writeString(nUsername);
				stream.Flush();

				string flap = string.Format("toc2_login {0} {1} {2} {3} English {4} {5}",
					authorizerServer,
					authorizerPort,
					nUsername,
					roast(password),
					REVISION,
					magicNumber(nUsername,password)					
					);

				sendFlap((byte)FrameType.Data, flap);


                //if login fails TOC drops Client's connection
				//else TOC sends "toc_signon" message
				flap = readFlap();

				if (flap.Equals("SIGN_ON:TOC2.0"))
				{
					sendFlap((byte)FrameType.Data, "toc_init_done");
					isConnected = true;

					//fireConnected();
					//setCaps(Cap.Talk+" "+Cap.AimIcq);
					return true;
				}
				else if (flap.StartsWith("ERROR"))
				{
					SignOff();
					closeConnection();
				}
			}
			catch(Exception e)
			{
				Console.WriteLine(e.Message);
			}

			return false;
		}
		

		
		/// <summary>
		/// sends a FLAP to the server
		/// </summary>
		/// <param name="frameType">type of FLAP</param>
		/// <param name="data">data to send</param>
		/// <seealso cref="FrameType"/>
		protected void sendFlap(byte frameType, string data)
		{
			//FLAP header
			stream.WriteByte(42); // *
			stream.WriteByte(frameType);
			writeShort((short)seqNo++);
			writeShort((short)(data.Length+1));
			//the data to send
			writeString(data);
			stream.WriteByte(0); //null terminating 0
			stream.Flush();
		}


		/// <summary>
		/// reads incoming FLAPs from the stream
		/// </summary>
		/// <returns>data string from FLAP</returns>
		private string readFlap()
		{
			string flap = null;

			if (stream.ReadByte() != '*')
				return null;

			stream.ReadByte(); //1 byte - frame type
			stream.ReadByte(); //2 bytes - seqNo 
			stream.ReadByte();

			int length = readShort(); //length of data

			byte[] buffer = new byte[length];
			stream.Read(buffer, 0, length); //read remaining data
			
			flap = Enc.GetString(buffer, 0, length);

			return flap;
		}

			
		//
		private byte[] getBytes(string s) 
		{
			return Enc.GetBytes(s);
		}


		//
		private void writeShort(short s)
		{
			stream.WriteByte((byte)((s >> 8) & 0xff));
			stream.WriteByte((byte)(s & 0xff));
		}


		//
		private void writeString(string s)
		{
			byte[] buffer = getBytes(s);
			stream.Write(buffer, 0, buffer.Length);
		}


		//
		private short readShort()
		{
			return (short)((stream.ReadByte()*0x100)+stream.ReadByte());

			//byte a = (byte)stream.ReadByte();
			//byte b = (byte)stream.ReadByte();
			//return (short)((b<<8) + a);
		}


		/// <summary>
		/// Normalizes a string
		/// </summary>
		/// <param name="s">string</param>
		/// <returns>normalized string</returns>
		protected string normalizeString(string s)
		{
			string normalized = "";

			for(int i=0; i<s.Length; i++)
			{
				if (s[i] != ' ')
				{
					normalized += Char.ToLower(s[i]);
				}
			}
			return normalized;
		}


		/*
		 * Roasts the given string with "Tic/Toc".
		 * Roasting, a weak form of encrypting the string
		 * 
		 * @param pass the string to roast (password)
		 * @return roasted string in hex format
		 */
		private String roast(String pass)
		{
			string roastString = "Tic/Toc";
			string temp = "";
			string s = "0x";
			for (int i=0;i<pass.Length;i++)
			{
				temp = string.Format("{0:X}",pass[i] ^ (byte)roastString[i%7]);
				if (temp.Length < 2)
					temp = "0"+temp;
				s = s + temp;
			}
			return s;
		}

		
		/*
		 * TOC2 magic number
		 * 
		 * @param user username of account to be signed on
		 * @param pass password of account to be signed on
		 * @return magic number
		 */
		private int magicNumber(String user, String pass)
		{
			int sn = user[0] - 96;
			int pw = pass[0] - 96;
			int a = sn * 7696 + 738816;
			int b = sn * 746512;
			int c = pw * a;
			return c - a + b + 71665152;
		}


		/// <summary>
		/// Adds a TOCListener to receive events from this connection
		/// </summary>
		/// <param name="listener">TOCListener</param>
		/// <seealso cref="TOCListener"/>
		public void AddTOCListener(TOCListener listener)
		{
			listeners.Add(listener);
		}


		/// <summary>
		/// Removes a TOCListener from receiving events
		/// </summary>
		/// <param name="listener">TOCListener</param>
		/// <seealso cref="TOCListener"/>
		public void RemoveTOCListener(TOCListener listener)
		{
			listeners.Remove(listener);
		}


		/// <summary>
		/// Propose a rendezvous with a buddy
		/// Not 100% positive how this command actually works over TOC2, 
		/// its functionality is unverified
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		/// <param name="uuid">capability UUID</param>
		/// <param name="cookie">cookie</param>
		/// <param name="seq">sequence number</param>
		/// <param name="rip">rendezvous IP, where you should connect</param>
		/// <param name="pip">buddy's proposer IP</param>
		/// <param name="vip">buddy's verified IP</param>
		/// <param name="port">port to connect on</param>
		/// <param name="tlv">base64 encoded values</param>
		/// <seealso cref="TOCConstants"/>
		public void ProposeRvous(string buddyName, string uuid, string cookie, string seq, string rip,
			string pip, string vip, int port, string tlv)
		{
			String data = "RVOUS_PROPOSE "+normalizeString(buddyName)+" "+uuid+" "+cookie+" "+seq+" "+rip+" "+
				pip+" "+vip+" "+port+" "+tlv;
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Cancels a rendezvous proposal
		/// </summary>
		/// <param name="buddyName">buddy's name</param>
		/// <param name="cookie">cookie</param>
		/// <param name="uuid">capability UUID</param>
		/// <param name="tlv">base64 encoded values</param>
		public void CancelRvous(string buddyName, string cookie, string uuid, string tlv)
		{
			string data = "toc_rvous_cancel "+normalizeString(buddyName)+" "+cookie+" "+uuid+" "+tlv;        
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Accepts a rendzevous from a buddy
		/// </summary>
		/// <param name="buddyName">buddy who you accepted proposal from</param>
		/// <param name="cookie">cookie</param>
		/// <param name="uuid">capability UUID</param>
		/// <param name="tlv">base64 encoded values</param>
		public void AcceptRvous(string buddyName, string cookie, string uuid, string tlv)
		{
			string data = "toc_rvous_accept "+normalizeString(buddyName)+" "+cookie+" "+uuid+" "+tlv;
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Change user's password
		/// </summary>
		/// <param name="oldPass">old password</param>
		/// <param name="newPass">new password</param>
		public void ChangePassword(string oldPass, string newPass)
		{
			string data = "toc_change_passwd "+oldPass+" "+newPass;
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Changes how a user's name is formatted. (such as 
		///	added spaces, capitals, etc..)
		/// </summary>
		/// <param name="name">new name format</param>
		public void SetNameFormat(string name)
		{
			string data = "toc_format_nickname \""+name+"\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Removes a buddy from the permit list
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		public void PermitDrop(string buddyName)
		{
			string data = "toc2_remove_permit "+normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Removes an array of buddies from the permit list
		/// </summary>
		/// <param name="buddyNames">string array of buddy names</param>
		public void PermitDrop(string[] buddyNames)
		{
			string data = "toc2_remove_permit";
			for(int i=0; i<buddyNames.Length; i++)
				data += " "+normalizeString(buddyNames[i]);

			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Removes a buddy from the deny/block list
		/// </summary>
		/// <param name="buddyName"></param>
		public void DenyDrop(string buddyName)
		{
			string data = "toc2_remove_deny "+normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Removes an array of buddies from the deny/block list
		/// </summary>
		/// <param name="buddyNames"string array of buddy names></param>
		public void DenyDrop(string[] buddyNames)
		{
			string data = "toc2_remove_deny";
			for(int i=0; i<buddyNames.Length; i++)
				data += " "+normalizeString(buddyNames[i]);

			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Add buddy to permit list
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		public void PutPermit(string buddyName)
		{
			string data = "toc2_add_permit "+normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Add an array of buddies to permit list
		/// </summary>
		/// <param name="buddyNames">array of buddy names</param>
		public void PutPermit(string[] buddyNames)
		{
			string data = "toc2_add_permit";
			for(int i=0; i<buddyNames.Length; i++)
				data += " "+normalizeString(buddyNames[i]);

			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Add buddy to deny/block list
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		public void PutDeny(string buddyName)
		{
			string data = "toc2_add_deny "+normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Add an array of buddy names to deny/block list
		/// </summary>
		/// <param name="buddyNames">array of buddy names</param>
		public void PutDeny(string[] buddyNames)
		{
			string data = "toc2_add_deny";
			for(int i=0; i<buddyNames.Length; i++)
				data += " "+normalizeString(buddyNames[i]);

			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Deletes a group from the buddy list
		/// </summary>
		/// <param name="groupName">group name</param>
		public void DropGroup(string groupName)
		{
			string data = "toc2_del_group \""+groupName+"\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Creates a new group in the buddy list
		/// </summary>
		/// <param name="groupName">group name</param>
		public void CreateGroup(string groupName)
		{
			string data = "toc2_new_group \""+groupName+"\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Search for a user based on certain criteria
		/// Server will respond with an error or goto_url
		/// </summary>
		/// <param name="first">first name</param>
		/// <param name="middle">middle name</param>
		/// <param name="last">last name</param>
		/// <param name="maiden">maiden name</param>
		/// <param name="city">city</param>
		/// <param name="state">state</param>
		/// <param name="country">country</param>
		/// <param name="email">email address</param>
		/// <param name="webSearch">web search</param>
		/// <seealso cref="TOCListener.UrlReceived()"/>
		public void SearchDirectory(string first, string middle, string last, string maiden,
			string city, string state, string country, string email, string webSearch)
		{
			if (first == null) first = "";
			if (middle == null) middle = "";
			if (last == null) last = "";
			if (maiden == null) maiden = "";
			if (city == null) city = "";
			if (state == null) state = "";
			if (country == null) country = "";
			if (email == null) email = "";
			if (webSearch == null) webSearch = "";

			string data = "toc_dir_search \"" + first + "\":\"" + middle + "\":\"" + last + 
				"\":\"" + maiden + "\":\"" + city + "\":\"" + state + "\":\"" + 
				country + "\":\"" + email + "\":\"" + webSearch + "\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Set user's directory info
		/// Server responds with dir_status
		/// </summary>
		/// <param name="first">first name</param>
		/// <param name="middle">middle name</param>
		/// <param name="last">last name</param>
		/// <param name="maiden">maiden name</param>
		/// <param name="city">city</param>
		/// <param name="state">state</param>
		/// <param name="country">country</param>
		/// <param name="email">email address</param>
		/// <param name="webSearch">allows for directory info to be searched via the web</param>
		/// <seealso cref="TOCListener.DirStatus()"/>
		public void SetDirectoryInfo(string first, string middle, string last, string maiden,
			string city, string state, string country, string email, string webSearch)
		{
			if (first == null) first = "";
			if (middle == null) middle = "";
			if (last == null) last = "";
			if (maiden == null) maiden = "";
			if (city == null) city = "";
			if (state == null) state = "";
			if (country == null) country = "";
			if (email == null) email = "";
			if (webSearch == null) webSearch = "";

			string data = "toc_set_dir \"" + first + "\":\"" + middle + "\":\"" + last + 
				"\":\"" + maiden + "\":\"" + city + "\":\"" + state + "\":\"" + 
				country + "\":\"" + email + "\":\"" + webSearch + "\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Retrieve's a buddy's directory info.
		/// Server responds with a goto_url or error
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		/// <seealso cref="TOCListener.UrlReceived()"/>
		public void GetDirectoryInfo(string buddyName)
		{
			string data = "toc_get_dir " + normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Retrieve buddy's profile info.
		/// Server reponds with goto_url or error
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		/// <seealso cref="TOCListener.UrlReceived()"/>
		public void GetBuddyInfo(string buddyName)
		{
			string data = "toc_get_info "+normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Set user profile info
		/// </summary>
		/// <param name="info"></param>
		public void SetProfile(string info)
		{
			string data = "toc_set_info \""+info+"\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Return from idle state
		/// </summary>
		public void ReturnFromIdle()
		{
			string data = "toc_set_idle 0";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Set client idle
		/// This only needs to be sent once, the server 
		/// handles incrementing the time from that point on
		/// </summary>
		/// <param name="seconds">seconds the client has been idle so far</param>
		public void SetIdle(int seconds)
		{
			string data = "toc_set_idle "+seconds;
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Accept a chat invitation
		/// </summary>
		/// <param name="roomID">id of room to join</param>
		public void AcceptChatInvite(int roomID)
		{
			string data = "toc_chat_accept "+roomID;
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Disconnect from a chat
		/// </summary>
		/// <param name="roomID">id of room to leave</param>
		public void DropChat(int roomID)
		{
			string data = "toc_chat_leave "+roomID;
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// warn a buddy in the chatroom
		/// </summary>
		/// <param name="roomID">room ID</param>
		/// <param name="buddyName">buddy to warn</param>
		/// <param name="anonymous">T/F to be anonymous</param>
		public void SendChatWarning(int roomID, string buddyName, bool anonymous)
		{
			string from = "norm";
			if (anonymous)
				from = "anon";
            
			String data = "toc_chat_evil " + roomID + normalizeString(buddyName) + from;

			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Whisper to a buddy, send private message
		/// to only that buddy within the chat
		/// </summary>
		/// <param name="roomID">room ID</param>
		/// <param name="buddyName">buddy name</param>
		/// <param name="message">message</param>
		public void SendChatWhisper(int roomID, string buddyName, string message)
		{
			string data = "toc_chat_whisper " + roomID + normalizeString(buddyName) + "\"" + message+"\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Send a message into a chatroom
		/// </summary>
		/// <param name="roomID">room ID</param>
		/// <param name="message">message</param>
		public void SendChatMessage(int roomID, string message)
		{
			string data = "toc_chat_send " + roomID + " \""+message+"\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Create a chatroom, name cannot contain spaces
		/// Server replies with a chat_join followed by
		/// chat_update_buddy pertaining to the user which
		/// created the chatroom
		/// </summary>
		/// <param name="name">chatroom name</param>
		/// <seealso cref="TOCListener.ChatroomJoined()"/>
		/// <seealso cref="TOCListener.ChatroomUpdate()"/>
		public void CreateChatRoom(String name)
		{
			string data = "toc_chat_join 4 "+name;
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Invite a buddy to join a chatroom
		/// </summary>
		/// <param name="roomId">room ID</param>
		/// <param name="buddyName">buddy name</param>
		/// <param name="message">message</param>
		public void SendChatInvite(string roomId, string buddyName, string message)
		{
			if (message == null) message = "";
			string data = "toc_chat_invite "+roomId+" \""+message+"\" "+normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Remove a buddy from buddy list
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		/// <param name="group">group</param>
		public void DropBuddy(string buddyName, string group)
		{
			string data = "toc2_remove_buddy "+normalizeString(buddyName)+" \""+group+"\"";
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Asks server to resend buddy status info, 
		/// responds with update_buddy which is parsed
		/// to see if buddy is online/offline/away/etc....
		/// Probably never have to call this unless
		/// client becomes out of sync for whatever
		/// reason.(which shouldn't happen at all)
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		public void GetBuddyStatus(string buddyName)
		{
			string data = "toc_get_status "+normalizeString(buddyName);
			sendFlap((byte)FrameType.Data, data+"\0");
		}


		/// <summary>
		/// Add buddy to buddy list
		/// </summary>
		/// <param name="buddyName">buddy name</param>
		/// <param name="group">group</param>
		public void AddBuddy(string buddyName, string group)
		{
			string data = "toc2_new_buddies {g:"+group+"\nb:"+buddyName+"\n}";
			sendFlap((byte)FrameType.Data, data+"\0");
			TOCBuddy buddy = new TOCBuddy(buddyName, group);
			buddies.Add(normalizeString(buddyName), buddy);
		}




		// ===================================================================
		// ===================================================================
		// ===================================================================
		//
		//
		// FIRE EVENTS BELOW HERE
		//
		//
		// ===================================================================
		// ===================================================================
		// ===================================================================







		/// <summary>
		/// 
		/// </summary>
		private void fireConfigReceived()
		{			
			foreach(TOCListener listener in listeners)
			{
				listener.BuddyListReceived(GetBuddies());
			}
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="message"></param>
		/// <param name="isBuddy"></param>
		/// <param name="auto"></param>
		private void fireMessageReceived(string buddy, string message, bool isBuddy, bool auto)
		{
			foreach(TOCListener listener in listeners)
			{
				listener.MessageReceived(buddy, message, isBuddy, auto);
			}
		}


		
		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="message"></param>
		/// <param name="isBuddy"></param>
		/// <param name="auto"></param>
		/// <param name="language"></param>
		private void fireMessageReceived(string buddy, string message, bool isBuddy, bool auto, string language)
		{
			foreach(TOCListener listener in listeners)
			{
				listener.MessageReceived(buddy, message, isBuddy, auto, language);
			}
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="status"></param>
		private void fireClientEvent(string buddy, int status)
		{
			foreach(TOCListener listener in listeners)
			{
				listener.ClientEvent(buddy, status);
			}
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="unknown"></param>
		private void fireBart(string buddy, string unknown)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddyArt(buddy, unknown);
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireBuddyAdded(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddyAdded(buddy);
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="group"></param>
		private void fireBuddyAdded(string buddy, string group)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddyAdded(buddy, group);
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireDenyAdded(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.DenyAdded(buddy);
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void firePermitAdded(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.PermitAdded(buddy);
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireBuddySignedOn(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddySignedOn(buddy);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireBuddyAway(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddyAway(buddy);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireBuddyReturned(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddyReturned(buddy);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireBuddySignedOff(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddySignedOff(buddy);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="error"></param>
		/// <param name="args"></param>
		private void fireTocError(int error, string[] args)
		{
			foreach(TOCListener listener in listeners)
				listener.Error(error, args);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="amount"></param>
		private void fireWarning(string buddy, int amount)
		{
			foreach(TOCListener listener in listeners)
				listener.Warned(buddy, amount);
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="roomID"></param>
		/// <param name="roomName"></param>
		private void fireChatJoined(string roomID, string roomName)
		{
			foreach(TOCListener listener in listeners)
				listener.ChatroomJoined(roomID, roomName);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="roomID"></param>
		/// <param name="sender"></param>
		/// <param name="message"></param>
		/// <param name="whisper"></param>
		private void fireChatMessageReceived(string roomID, string sender, string message, bool whisper)
		{
			foreach(TOCListener listener in listeners)
				listener.ChatroomMessageReceived(roomID, sender, message, whisper);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="roomID"></param>
		/// <param name="sender"></param>
		/// <param name="message"></param>
		/// <param name="whisper"></param>
		/// <param name="language"></param>
		/// <param name="unknown"></param>
		private void fireChatMessageReceived(string roomID, string sender, string message, bool whisper, string language, string unknown)
		{
			foreach(TOCListener listener in listeners)
				listener.ChatroomMessageReceived(roomID, sender, message, whisper, language, unknown);
		}
			



		/// <summary>
		/// 
		/// </summary>
		/// <param name="roomID"></param>
		/// <param name="inside"></param>
		/// <param name="buddy"></param>
		private void fireChatUpdate(string roomID, bool inside, string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.ChatroomUpdate(roomID, buddy, inside);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="roomID"></param>
		/// <param name="roomName"></param>
		/// <param name="sender"></param>
		/// <param name="message"></param>
		private void fireChatInvite(string roomID, string roomName, string sender, string message)
		{
			foreach(TOCListener listener in listeners)
				listener.ChatroomInvite(roomID, roomName, sender, message);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="roomID"></param>
		private void fireChatDropped(string roomID)
		{
			foreach(TOCListener listener in listeners)
				listener.ChatroomDropped(roomID);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="url"></param>
		private void fireUrlReceived(string url)
		{
			foreach(TOCListener listener in listeners)
				listener.UrlReceived(url);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="code"></param>
		/// <param name="args"></param>
		private void fireDirStatus(int code, string[] args)
		{
			foreach(TOCListener listener in listeners)
				listener.DirStatus(code, args);
		}




		/// <summary>
		/// 
		/// </summary>
		/// <param name="code"></param>
		/// <param name="args"></param>
		private void fireNickStatus(int code, string[] args)
		{
			foreach(TOCListener listener in listeners)
				listener.NickStatus(code, args);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="code"></param>
		/// <param name="args"></param>
		private void firePasswordStatus(int code, string[] args)
		{
			foreach(TOCListener listener in listeners)
				listener.AdminPasswordStatus(code, args);
		}


		
		/// <summary>
		/// 
		/// </summary>
		/// <param name="isPaused"></param>
		private void firePause(bool isPaused)
		{
			foreach(TOCListener listener in listeners)
				listener.IsPaused(isPaused);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="UUID"></param>
		/// <param name="cookie"></param>
		/// <param name="seq"></param>
		/// <param name="rip"></param>
		/// <param name="pip"></param>
		/// <param name="vip"></param>
		/// <param name="port"></param>
		/// <param name="tlv"></param>
		private void fireRvousPropose(string buddy, string UUID, string cookie, int seq, string rip, 
			string pip, string vip, string port, string[] tlv)
		{
			foreach(TOCListener listener in listeners)
				listener.RvousPropose(buddy, UUID, cookie, seq, rip, pip, vip, port, tlv);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		/// <param name="group"></param>
		private void fireBuddyRemoved(string buddy, string group)
		{
			foreach(TOCListener listener in listeners)
				listener.BuddyDropped(buddy, group);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireBuddyRemovedFromDeny(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.DenyDropped(buddy);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="buddy"></param>
		private void fireBuddyRemovedFromPermit(string buddy)
		{
			foreach(TOCListener listener in listeners)
				listener.PermitDropped(buddy);
		}



		/// <summary>
		/// 
		/// </summary>
		/// <param name="group"></param>
		private void fireGroupRemoved(string group)
		{
			foreach(TOCListener listener in listeners)
				listener.GroupDropped(group);
		}


	}
}